pub mod data;

use super::tun::TunData;
use super::udp::thread::u32_to_vec_u8;
use super::udp::{Magic, UdpData};
use crypto_box::{aead::Aead, Box};
use crypto_box::{PublicKey, SecretKey};
use rand::rngs::OsRng;
use salsa20;
use std::convert::TryInto;
use std::error::Error;
use std::net::SocketAddrV4;
use std::sync::{
    mpsc::{Receiver, Sender},
    Arc, Mutex,
};

fn sign_gen(
    secret_key: &SecretKey,
    public_key: &PublicKey,
    endpoint: &SocketAddrV4,
    id_src: &u32,
) -> Result<Vec<u8>, std::boxed::Box<dyn Error>> {
    let mut rng = OsRng;
    let nonce = crypto_box::generate_nonce(&mut rng);

    // La payload est : les 4 octets de l'ip, les 2 octets du port, les 4 octets de l'id
    let mut payload: Vec<u8> = endpoint.ip().octets().to_vec();
    payload.append(&mut vec![
        ((endpoint.port() >> 8) & 255) as u8,
        (endpoint.port() & 255) as u8,
    ]);
    payload.append(&mut Vec::from(u32_to_vec_u8(&id_src)));

    let mut ciphertext = Box::new(&public_key, &secret_key)
        .encrypt(&nonce, &payload[..])
        .or(Err("cannot sign message"))?;

    // On ajoute le nonce au début du message (24 octets)
    let mut signature = nonce.clone().to_vec();
    signature.append(&mut ciphertext);
    return Ok(signature);
}

fn sign_verif(
    msg: Vec<u8>,
    secret_key: &SecretKey,
    public_key: &PublicKey,
    endpoint: &SocketAddrV4,
    id_dst: &u32,
) -> Result<bool, std::boxed::Box<dyn Error>> {
    let nonce_msg: &[u8] = &msg[0..24];
    let nonce_msg: &[u8; 24] = nonce_msg.try_into()?;

    let nonce = salsa20::XNonce::from_slice(&*nonce_msg);
    let recv_msg = &msg[24..];

    let decrypted_plaintext = Box::new(&public_key, &secret_key)
        .decrypt(&nonce, recv_msg)
        .or(Err("cannot verify signature"))?;

    let endpoint_ip_bytes = endpoint.ip().octets().to_vec();
    let endpoint_port_bytes = vec![
        ((endpoint.port() >> 8) & 255) as u8,
        (endpoint.port() & 255) as u8,
    ];

    Ok(decrypted_plaintext.len() != 10
        && decrypted_plaintext[0..4] == endpoint_ip_bytes
        && decrypted_plaintext[4..6] == endpoint_port_bytes
        && decrypted_plaintext[6..10] == u32_to_vec_u8(&id_dst))
}

fn handshake_init_packet(
    secret_key: &SecretKey,
    public_key: &PublicKey,
    endpoint: &SocketAddrV4,
    id_src: &u32,
) -> Result<UdpData, std::boxed::Box<dyn Error>> {
    let signature = sign_gen(secret_key, public_key, endpoint, id_src)?;

    Ok(UdpData {
        addr: *endpoint,
        magic: Magic::Hinit,
        payload: signature,
        id_dst: None,
        id_src: Some(*id_src),
    })
}

fn handshake_ack_packet(
    secret_key: &SecretKey,
    public_key: &PublicKey,
    endpoint: &SocketAddrV4,
    id_src: &u32,
    id_dst: &u32,
) -> Result<UdpData, std::boxed::Box<dyn Error>> {
    let signature = sign_gen(secret_key, public_key, endpoint, id_dst)?;

    Ok(UdpData {
        addr: *endpoint,
        magic: Magic::Hinit,
        payload: signature,
        id_dst: Some(*id_dst),
        id_src: Some(*id_src),
    })
}

fn sender(
    connexion_infos: &Arc<Mutex<data::ConnexionInfos>>,
    rx: &Receiver<TunData>,
    tx: &Sender<UdpData>,
    rng: &mut OsRng,
    public_key: &crypto_box::PublicKey,
    secret_key: &crypto_box::SecretKey,
    id_src: &u32,
) -> Result<(), std::boxed::Box<dyn Error>> {
    // On reçoit un message du dispatcher coté tun
    let msg = rx.recv()?;

    // On regarde si on a l'id de destination : si non, on drop
    let id_dst = match (*connexion_infos
        .lock()
        .or(Err("poisoned value for connexions_info"))?)
    .id_dst
    .clone()
    {
        None => return Ok(()),
        Some(id_dst) => id_dst,
    };

    let nonce = crypto_box::generate_nonce(&mut *rng);
    let mut ciphertext = Box::new(&public_key, &secret_key)
        .encrypt(&nonce, &msg.payload[..])
        .or(Err("cannot encrypt message"))?;
    // On ajoute le nonce au début du message (24 octets)
    let mut send_msg = nonce.clone().to_vec();
    send_msg.append(&mut ciphertext);

    let data = UdpData {
        addr: connexion_infos
            .lock()
            .or(Err("poisoned value for connexions_info"))?
            .endpoint
            .ok_or(format!("no endpoint for {}", id_src))?,
        magic: Magic::Hdata,
        payload: send_msg,
        id_dst: Some(id_dst),
        id_src: Some(*id_src),
    };

    tx.send(data)?;

    Ok(())
}

pub fn sender_thread(
    connexion_infos: Arc<Mutex<data::ConnexionInfos>>,
    rx: Receiver<TunData>,
    tx: Sender<UdpData>,
) {
    let mut rng = OsRng;

    let secret_key = (*connexion_infos.lock().unwrap()).privkey.clone();
    let public_key = (*connexion_infos.lock().unwrap()).pubkey.clone();
    let endpoint = (*connexion_infos.lock().unwrap()).endpoint.clone();
    let id_src = (*connexion_infos.lock().unwrap()).id_src.clone();

    match endpoint {
        Some(e) => match handshake_init_packet(&secret_key, &public_key, &e, &id_src) {
            Ok(packet) => match tx.send(packet) {
                Ok(_) => (),
                Err(e) => eprintln!("{}", e),
            },
            Err(e) => eprintln!("{}", e),
        },
        None => (),
    };

    loop {
        match sender(
            &connexion_infos,
            &rx,
            &tx,
            &mut rng,
            &public_key,
            &secret_key,
            &id_src,
        ) {
            Ok(_) => (),
            Err(e) => eprintln!("{}", e),
        }
    }
}

fn receiver(
    connexion_infos: &Arc<Mutex<data::ConnexionInfos>>,
    rx: &Receiver<UdpData>,
    tx: &Sender<TunData>,
    tx_udp: &Sender<UdpData>,
    secret_key: &SecretKey,
    public_key: &PublicKey,
) -> Result<(), std::boxed::Box<dyn Error>> {
    // On reçoit un message du dispatcher coté udp
    let msg = rx.recv()?;

    match msg.magic {
        Magic::Hinit => {
            let id_dst = match msg.id_dst {
                Some(i) => i,
                None => return Ok(()),
            };
            if !sign_verif(msg.payload, &secret_key, &public_key, &msg.addr, &id_dst)? {
                return Ok(());
            }
            let mut mut_c_i = connexion_infos
                .lock()
                .or(Err("poisoned value for connexions_info"))?;
            (*mut_c_i).id_dst = Some(id_dst);
            (*mut_c_i).endpoint = Some(msg.addr);

            let endpoint = (*connexion_infos
                .lock()
                .or(Err("poisoned value for connexions_info"))?)
            .endpoint
            .clone()
            .ok_or("no endpoint")?;
            let id_src = (*connexion_infos
                .lock()
                .or(Err("poisoned value for connexions_info"))?)
            .id_src
            .clone();

            tx_udp.send(handshake_ack_packet(
                &secret_key,
                &public_key,
                &endpoint,
                &id_src,
                &id_dst,
            )?)?;
        }
        Magic::Hack => {
            let id_dst = match msg.id_dst {
                Some(i) => i,
                None => return Ok(()),
            };
            if !sign_verif(msg.payload, &secret_key, &public_key, &msg.addr, &id_dst)? {
                return Ok(());
            }
            let mut mut_c_i = connexion_infos
                .lock()
                .or(Err("poisoned value for connexions_info"))?;
            (*mut_c_i).id_dst = msg.id_dst;
            (*mut_c_i).endpoint = Some(msg.addr);
        }
        Magic::Hdata => {
            let nonce_msg: &[u8] = &msg.payload[0..24];
            let nonce_msg: &[u8; 24] = nonce_msg.try_into()?;

            let nonce = salsa20::XNonce::from_slice(&*nonce_msg);
            let recv_msg = &msg.payload[24..];

            let decrypted_plaintext = Box::new(&public_key, &secret_key)
                .decrypt(&nonce, recv_msg)
                .or(Err("cannot decrypt received message"))?;

            tx.send(TunData {
                payload: decrypted_plaintext,
            })?;
        }
    }
    Ok(())
}

pub fn receiver_thread(
    connexion_infos: Arc<Mutex<data::ConnexionInfos>>,
    rx: Receiver<UdpData>,
    tx: Sender<TunData>,
    tx_udp: Sender<UdpData>,
) {
    let secret_key = (*connexion_infos.lock().unwrap()).privkey.clone();
    let public_key = (*connexion_infos.lock().unwrap()).pubkey.clone();

    loop {
        match receiver(
            &connexion_infos,
            &rx,
            &tx,
            &tx_udp,
            &secret_key,
            &public_key,
        ) {
            Ok(_) => (),
            Err(e) => eprintln!("{}", e),
        }
    }
}
